package be.iminds.ilabt.jfed.preferences;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.*;
import be.iminds.ilabt.jfed.git.GitAuthPreferences;
import be.iminds.ilabt.jfed.git.SingleSshGitAuthPreferences;
import be.iminds.ilabt.jfed.git.SingleUserPassGitAuthPreferences;
import be.iminds.ilabt.jfed.lowlevel.testbed_info.ApiInfo;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.lib.RetrySettings;
import be.iminds.ilabt.jfed.lowlevel.testbed_info.TestbedInfoSource;
import be.iminds.ilabt.jfed.lowlevel.connection.ConnectionBuilder;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.InvalidSshKeyInfoException;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.PuTTYFilesKeyInfo;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.pref_ssh_key_info.PreferencesSshKeyInfoFactory;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import be.iminds.ilabt.jfed.util.common.IOUtils;
import be.iminds.ilabt.jfed.util.library.JFedUtils;
import be.iminds.ilabt.jfed.util.common.TextUtil;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.PropertiesConfiguration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

/**
 * JFedCorePreferences
 */
public class JFedCorePreferences implements JFedPreferences {
    private final static Logger LOG = LoggerFactory.getLogger(JFedCorePreferences.class);
    @Nullable protected final File propertiesFile;
    protected final PropertiesConfiguration config;

    public JFedCorePreferences() {
        this.propertiesFile = JFedUtils.createPreferencesFile("experimenter-ssh.properties");

        if (propertiesFile != null) {
            try {
                config = new PropertiesConfiguration();
                config.setDelimiterParsingDisabled(true);
                config.setFile(propertiesFile);
                config.load();
            } catch (ConfigurationException e) {
                LOG.error("Could not create PropertiesConfiguration", e);
                throw new RuntimeException("Could not create PropertiesConfiguration", e);
            }
        } else {
            LOG.error("Fallback: Settings will NOT be saved!");
            config = new PropertiesConfiguration();
            config.setDelimiterParsingDisabled(true);
        }

        //no autosave! config.setAutoSave(true);
        config.setThrowExceptionOnMissing(false);

        //use defaults to fill in blanks
        fillinBlanks(config);
    }
    protected JFedCorePreferences(File propertiesFile, PropertiesConfiguration config) {
        this.propertiesFile = propertiesFile;
        this.config = config;

        //no autosave! config.setAutoSave(true);
        config.setThrowExceptionOnMissing(false);

        //use defaults to fill in blanks
        fillinBlanks(config);
    }
    public JFedCorePreferences makeNoSaveCopy() {
        PropertiesConfiguration config = new PropertiesConfiguration();
        config.copy(this.config);
        return new JFedCorePreferences(null, config);
    }

    public void fillinBlanks() {
        fillinBlanks(config);
    }

    protected void fillinBlanks(PropertiesConfiguration config) {
        if (!config.containsKey(CorePreferenceKey.PREF_TESTBEDS_JSON_SOURCE.key))
            config.setProperty(CorePreferenceKey.PREF_TESTBEDS_JSON_SOURCE.key, "AUTO");

        if (!config.containsKey(CorePreferenceKey.PREF_GUICONFIG_SOURCE.key))
            config.setProperty(CorePreferenceKey.PREF_GUICONFIG_SOURCE.key, "AUTO");

        if (!config.containsKey(CorePreferenceKey.PREF_DEBUG_CPU.key))
            config.setProperty(CorePreferenceKey.PREF_DEBUG_CPU.key, false);

        if (!config.containsKey(CorePreferenceKey.PREF_LOGGING_OVERWRITE.key))
            config.setProperty(CorePreferenceKey.PREF_LOGGING_OVERWRITE.key, false);
        if (!config.containsKey(CorePreferenceKey.PREF_LOGGING_ROOT_LEVEL.key))
            config.setProperty(CorePreferenceKey.PREF_LOGGING_ROOT_LEVEL.key, "INFO");
//        if (!config.containsKey(CorePreferenceKey.PREF_LOGGING_APPENDER_BUGREPORT_LEVEL.key))
//            config.setProperty(CorePreferenceKey.PREF_LOGGING_APPENDER_BUGREPORT_LEVEL.key, "INFO");
//        if (!config.containsKey(CorePreferenceKey.PREF_LOGGING_APPENDER_STDOUT_LEVEL.key))
//            config.setProperty(CorePreferenceKey.PREF_LOGGING_APPENDER_STDOUT_LEVEL.key, "INFO");
        if (!config.containsKey(CorePreferenceKey.PREF_LOGGING_EXTTERM_LEVEL.key))
            config.setProperty(CorePreferenceKey.PREF_LOGGING_EXTTERM_LEVEL.key, "DEBUG");

        if (!config.containsKey(CorePreferenceKey.PREF_TESTBEDSPECIFIC_EXOGENI_FORCE_EXOSM.key))
            config.setProperty(CorePreferenceKey.PREF_TESTBEDSPECIFIC_EXOGENI_FORCE_EXOSM.key, "AUTO");

        if (!config.containsKey(CorePreferenceKey.PREF_TESTBEDSPECIFIC_EXOGENI_RECOVER_METHOD.key))
            config.setProperty(CorePreferenceKey.PREF_TESTBEDSPECIFIC_EXOGENI_RECOVER_METHOD.key, "BOTH");

        if (!config.containsKey(CorePreferenceKey.PREF_SSHPROXY_USE_FOR_JFED.key))
            config.setProperty(CorePreferenceKey.PREF_SSHPROXY_USE_FOR_JFED.key, ProxyPreferencesManager.NEVER_USE_PROXY);

        if (!config.containsKey(CorePreferenceKey.PREF_SSHPROXY_USE_FOR_SSH.key))
            config.setProperty(CorePreferenceKey.PREF_SSHPROXY_USE_FOR_SSH.key, ProxyPreferencesManager.NEVER_USE_PROXY);

        if (!config.containsKey(CorePreferenceKey.PREF_SSHPROXY_DNS_OVER_PROXY_JFED.key))
            config.setProperty(CorePreferenceKey.PREF_SSHPROXY_DNS_OVER_PROXY_JFED.key, false);

        if (!config.containsKey(CorePreferenceKey.PREF_SSHPROXY_DNS_OVER_PROXY_SSH.key))
            config.setProperty(CorePreferenceKey.PREF_SSHPROXY_DNS_OVER_PROXY_JFED.key, false);

        if (!config.containsKey(CorePreferenceKey.PREF_GIT_AUTH_METHOD.key))
            config.setProperty(CorePreferenceKey.PREF_GIT_AUTH_METHOD.key, "SSH_LOGIN");

        if (!config.containsKey(CorePreferenceKey.PREF_SOCKET_CONNECT_TIMEOUT_MS.key))
            config.setProperty(CorePreferenceKey.PREF_SOCKET_CONNECT_TIMEOUT_MS.key, ConnectionBuilder.DEFAULT_SOCKET_CONNECT_TIMEOUT_MS);

        if (!config.containsKey(CorePreferenceKey.PREF_SSH_USE_LOGIN_KEY.key))
            config.setProperty(CorePreferenceKey.PREF_SSH_USE_LOGIN_KEY.key, true);

        if (!config.containsKey(CorePreferenceKey.PREF_SSH_USE_CUSTOM_KEY.key))
            config.setProperty(CorePreferenceKey.PREF_SSH_USE_CUSTOM_KEY.key, false);

        if (!config.containsKey(CorePreferenceKey.PREF_SOCKET_READ_TIMEOUT_MS.key))
            config.setProperty(CorePreferenceKey.PREF_SOCKET_READ_TIMEOUT_MS.key, ConnectionBuilder.DEFAULT_SOCKET_READ_TIMEOUT_MS);

        //make sure at least one ssh key option is true
        if (!getBoolean(CorePreferenceKey.PREF_SSH_USE_LOGIN_KEY, false)
                && !getBoolean(CorePreferenceKey.PREF_SSH_USE_CUSTOM_KEY, false))
            config.setProperty(CorePreferenceKey.PREF_SSH_USE_LOGIN_KEY.key, true);

//        if (getRetrySettings() == null)
//            setRetrySettings(new RetrySettings()/* = use internal default values*/);
        setRetrySettings(getRetrySettings()); //getRetrySettings() falls back to defaults, so set using them
    }

    public boolean containsPreference(@Nonnull PreferenceKey key) {
        return config.containsKey(key.getKey());
    }

    public boolean hasKey(@Nonnull PreferenceKey key) {
        return config.containsKey(key.getKey());
    }

    @Nonnull
    public String getString(@Nonnull PreferenceKey key, @Nonnull String defaultValue) {
        String res = getString(key);
        return res == null ? defaultValue : res;
    }

    @Override
    @Nullable
    public String getString(@Nonnull PreferenceKey key) {
        if (!config.containsKey(key.getKey())) return null;

        Object val = config.getProperty(key.getKey());
        if (val == null) {
            LOG.warn("BUG in JFedCorePreferences: existing preference for \"" + key.getKey() + "\" returned null. Using fallback to null.");
            return null;
        }
        if (val instanceof String)
            return (String) val;

        LOG.debug("Preference \"" + key.getKey() + "\" is of type " + val.getClass().getName() + " instead of String. Will convert it to String.");
        return val + "";
    }

    @Override
    @Nonnull
    public List<String> getStringList(@Nonnull PreferenceKey key) {
        if (!config.containsKey(key.getKey())) return Collections.emptyList();

        Object val = config.getProperty(key.getKey());
        if (val == null) {
            LOG.warn("BUG in JFedCorePreferences: existing preference for \"" + key.getKey() + "\" returned null. Using fallback to empty list.");
            return Collections.emptyList();
        }
        if (val instanceof String) {
            String s = (String) val;
            List<String> res = Arrays.asList(s.split(","));
            assert res.stream().noneMatch(Objects::isNull);
            return res;
        }
        if (val instanceof List) {
            //should not happen when config.setDelimiterParsingDisabled(true);   But we can handle it anyway.
            List<String> res = (List) val;
            assert res.stream().noneMatch(Objects::isNull);
            return res;
        }

        LOG.debug("Preference \"" + key.getKey() + "\" is of type " + val.getClass().getName() + " instead of String or String List. Will convert it to List with one String.");
        return Collections.singletonList(val + "");
    }

    @Override
    public void setString(@Nonnull PreferenceKey key, @Nonnull String value) {
        config.setProperty(key.getKey(), value);
    }
    @Override
    public void setStringList(@Nonnull PreferenceKey key, @Nonnull List<String> value) {
        assert value.stream().noneMatch(Objects::isNull);
        config.setProperty(key.getKey(), String.join(",", value));
    }

    @Nullable
    public Integer getInt(@Nonnull PreferenceKey key) {
        if (!config.containsKey(key.getKey())) return null;
        return config.getInt(key.getKey());
    }

    public int getInt(@Nonnull PreferenceKey key, int defaultValue) {
        if (!config.containsKey(key.getKey())) return defaultValue;
        return config.getInt(key.getKey());
    }

    public void setInt(@Nonnull PreferenceKey key, int value) {
        config.setProperty(key.getKey(), value);
    }

    public Boolean getBoolean(@Nonnull PreferenceKey key) {
        if (!config.containsKey(key.getKey())) return null;

        Object val = config.getProperty(key.getKey());
        assert val != null;

        Boolean res = TextUtil.objectToBoolean(val);
        if (res != null)
            return res;
        else
            throw new IllegalStateException("Stored preference for " + key.getKey() + " is null. This is not a valid boolean value.");
    }

    public boolean getBoolean(@Nonnull PreferenceKey key, boolean defaultValue) {
        Boolean res = getBoolean(key);
        return res == null ? defaultValue : res;
    }

    public void setBoolean(@Nonnull PreferenceKey key, boolean value) {
        config.setProperty(key.getKey(), value);
    }

    public File getFile(@Nonnull PreferenceKey key) {
        if (!config.containsKey(key.getKey())) return null;
        return new File(config.getString(key.getKey()));
    }

    public void setFile(@Nonnull PreferenceKey key, @Nonnull File value) {
        config.setProperty(key.getKey(), value.getPath());
    }

    public void clear(@Nonnull PreferenceKey key) {
        config.clearProperty(key.getKey());
    }

    @Nullable
    public File getPropertiesFile() {
        return propertiesFile;
    }


    @Override
    public void addPreferenceChangeListener(@Nonnull PreferenceKey prefDebugCpu, @Nonnull PreferenceChangeListener preferenceChangeListener) {
        //TODO: implement
    }


//  no longer needed: AuthorityFinder always checks the current preferences
//    public void updateForceExoSM(TestbedInfoSource testbedInfoSource) {


    @Override
    public void setRetrySettings(@Nonnull RetrySettings retrySettings) {
        setInt(CorePreferenceKey.RETRY_TIMEOUT_MAX, retrySettings.getTimeoutMaxRetries());
        setInt(CorePreferenceKey.RETRY_BUSY_MAX, retrySettings.getBusyMaxRetries());
        setInt(CorePreferenceKey.RETRY_POSTEMPERR_MAX, retrySettings.getPossibleTemporaryErrorMaxRetries());

        setString(CorePreferenceKey.RETRY_TIMEOUT_MS_DELAY, retrySettings.getTimeoutMsBeforeRetry() + "");
        setString(CorePreferenceKey.RETRY_BUSY_MS_DELAY, retrySettings.getBusyMsBeforeRetry() + "");
        setString(CorePreferenceKey.RETRY_POSTEMPERR_MS_DELAY, retrySettings.getPossibleTemporaryErrorMsBeforeRetry() + "");
    }

    @Nonnull
    @Override
    public RetrySettings getRetrySettings() {
        Integer timeoutMaxRetries =
                getInt(CorePreferenceKey.RETRY_TIMEOUT_MAX, RetrySettings.DEFAULT_TIMEOUT_MAX_RETRIES);
        Integer busyMaxRetries = getInt(CorePreferenceKey.RETRY_BUSY_MAX, RetrySettings.DEFAULT_BUSY_MAX_RETRIES);
        Integer possibleTemporaryErrorMaxRetries =
                getInt(CorePreferenceKey.RETRY_POSTEMPERR_MAX, RetrySettings.DEFAULT_TIMEOUT_MAX_RETRIES);

        String timeoutMsBeforeRetry = getString(CorePreferenceKey.RETRY_TIMEOUT_MS_DELAY,
                Integer.toString(RetrySettings.DEFAULT_TIMEOUT_MS_BEFORE_RETRY));
        String busyMsBeforeRetry = getString(CorePreferenceKey.RETRY_BUSY_MS_DELAY,
                Integer.toString(RetrySettings.DEFAULT_BUSY_MS_BEFORE_RETRY));
        String possibleTemporaryErrorMsBeforeRetry = getString(CorePreferenceKey.RETRY_POSTEMPERR_MS_DELAY,
                Integer.toString(RetrySettings.DEFAULT_TIMEOUT_MS_BEFORE_RETRY));

        try {
            return new RetrySettings(
                    timeoutMaxRetries, Long.parseLong(timeoutMsBeforeRetry),
                    busyMaxRetries, Long.parseLong(busyMsBeforeRetry),
                    possibleTemporaryErrorMaxRetries, Long.parseLong(possibleTemporaryErrorMsBeforeRetry)
            );
        } catch (NumberFormatException e) {
            LOG.error("Error parsing a RetrySettings", e);
            return new RetrySettings();
        }
    }


    public void setScs(Server server) {
        if (server == null) {
            clear(CorePreferenceKey.PREF_SCS_URN);
            clear(CorePreferenceKey.PREF_SCS_URL);
            return;
        }
        Service service = ApiInfo.findService(server, ApiInfo.ApiName.GENI_SCS, 1);
        if (service == null)
            throw new RuntimeException("Authority " + server.getDefaultComponentManagerUrn() + " has no SCS service");
        URL u;
        try {
            u = new URL(service.getUrl());
        } catch (MalformedURLException e) {
            throw new RuntimeException("Authority " + server.getDefaultComponentManagerUrn() + " has invalid SCS Url \"" + service.getUrl() + "\"", e);
        }
        try {
            setScs(new GeniUrn(service.getUrn()), u);
        } catch (GeniUrn.GeniUrnParseException e) {
            throw new RuntimeException("Authority " + server.getDefaultComponentManagerUrn() + " has invalid SCS Urn \"" + service.getUrn() + "\"");
        }
    }

    public void setScs(GeniUrn scsUrn, URL url) {
        if (scsUrn == null && url == null) {
            clear(CorePreferenceKey.PREF_SCS_URN);
            clear(CorePreferenceKey.PREF_SCS_URL);
            return;
        }
        if (scsUrn != null)
            setString(CorePreferenceKey.PREF_SCS_URN, scsUrn.toString());
        else
            clear(CorePreferenceKey.PREF_SCS_URN);
        setString(CorePreferenceKey.PREF_SCS_URL, url.toString());
    }

    public void setScs(URL url) {
        if (url == null) {
            clear(CorePreferenceKey.PREF_SCS_URN);
            clear(CorePreferenceKey.PREF_SCS_URL);
            return;
        }
        clear(CorePreferenceKey.PREF_SCS_URN);
        setString(CorePreferenceKey.PREF_SCS_URL, url.toString());
    }

    public URL getScsUrl() {
        String u = getString(CorePreferenceKey.PREF_SCS_URL);
        if (u == null) return null;
        try {
            return new URL(u);
        } catch (MalformedURLException e) {
            clear(CorePreferenceKey.PREF_SCS_URL);
            throw new RuntimeException("Stored SCS URL is not valid URL (and has been removed from config). value=\"" + u + "\"");
        }
    }

    public GeniUrn getScsUrn() {
        String u = getString(CorePreferenceKey.PREF_SCS_URN);
        if (u == null) return null;
        try {
            return new GeniUrn(u);
        } catch (GeniUrn.GeniUrnParseException e) {
            clear(CorePreferenceKey.PREF_SCS_URN);
            throw new RuntimeException("Stored SCS URN is not valid URN (and has been removed from config). value=\"" + u + "\"");
        }
    }

    @Nonnull
    public static Server createFakeScsServer(@Nonnull URL scsUrl, @Nullable GeniUrn scsUrn) {
        TestbedBuilder testbedBuilder = new TestbedBuilder();
        ServerBuilder serverBuilder = new ServerBuilder();
        ServiceBuilder serviceBuilder = new ServiceBuilder();

        testbedBuilder.addServerBuilder(serverBuilder);
        serverBuilder.addServiceBuilder(serviceBuilder);

        testbedBuilder.setLongName("Config specified SCS");
        serverBuilder.setName("Config specified SCS");

        if (scsUrn != null) {
            testbedBuilder.setDefaultComponentManagerUrn(scsUrn != null ? scsUrn.toString() : null);
            serverBuilder.setDefaultComponentManagerUrn(scsUrn.toString());
        }

        serviceBuilder.setUrn(scsUrn.toString());
        serviceBuilder.setUrl(scsUrl.toExternalForm());
        serviceBuilder.setApi(ApiInfo.ApiName.GENI_SCS.getId());
        serviceBuilder.setApiVersion("1");

        Testbed scsTestbed = testbedBuilder.create();
        Server scsServer = scsTestbed.getServers().get(0);
        return scsServer;
    }

    public Server getScsAuthority(TestbedInfoSource testbedInfoSource) {
        URL scsUrl = getScsUrl();
        String scsUrnString = getString(CorePreferenceKey.PREF_SCS_URN);
        GeniUrn scsUrn = scsUrnString == null ? null : GeniUrn.parse(scsUrnString);

        Server scsServer = null;
        if (scsUrn != null) {
            scsServer = testbedInfoSource.getByUrnExact(scsUrn);
            Service scsService = ApiInfo.findService(scsServer, ApiInfo.ApiName.GENI_SCS, 1);
            assert scsServer == null || (scsService != null && Objects.equals(scsService.getUrl(), scsUrl.toExternalForm()));
        }
        if (scsServer == null && scsUrl != null)
            scsServer = testbedInfoSource.getServerByUrl(scsUrl);
        if (scsServer == null && scsUrl != null) {
            scsServer = createFakeScsServer(scsUrl, scsUrn);
        }

        return scsServer;
    }

    @Nullable
    public GitAuthPreferences getGitAuthPreferences(@Nullable GeniUser user) {
        String gitAuthMethod = getString(CorePreferenceKey.PREF_GIT_AUTH_METHOD);
        if (gitAuthMethod == null) return null;

        switch (gitAuthMethod.toUpperCase()) {
            default: LOG.error("Unsupported preference for PREF_GIT_AUTH_METHOD: '"+gitAuthMethod+"'");
            case "NONE": return null;
            case "SSH_LOGIN": {
                //noinspection ConstantConditions
                if (user == null || user.getPrivateKey() == null)
                    return null;

                //use priv key in login PEM file
                return new SingleSshGitAuthPreferences(user.getPrivateKey());
            }
            case "SSH_CUSTOM": {
                String keyFileName = getString(CorePreferenceKey.PREF_GIT_AUTH_KEYFILE);
                if (keyFileName == null)
                    throw new RuntimeException("Git auth preferences not complete");
                String keyFileContent;
                try {
                    keyFileContent = IOUtils.fileToString(keyFileName);
                } catch (IOException e) {
                    throw new RuntimeException("Error reading private SSH key file '"+keyFileName+"'", e);
                }
                return new SingleSshGitAuthPreferences(keyFileContent);
            }
            case "USER_PASS": {
                String username = getString(CorePreferenceKey.PREF_GIT_AUTH_USERNAME);
                String pass = getString(CorePreferenceKey.PREF_GIT_AUTH_PASSWORD);
                return new SingleUserPassGitAuthPreferences(username, pass);
            }
        }
    }

    @Nullable
    public PuTTYFilesKeyInfo getCustomSshKeyInfo() {
        if (getBoolean(CorePreferenceKey.PREF_SSH_USE_CUSTOM_KEY, false)) {
            try {
                if (config.containsKey(CorePreferenceKey.PREF_SSH_PUTTY_KEY_FILE.getKey())) {
                    return PreferencesSshKeyInfoFactory.createPreferencesPuttySshKeyInfo(this, CorePreferenceKey.PREF_SSH_USE_CUSTOM_KEY,
                            CorePreferenceKey.PREF_SSH_PUTTY_KEY_FILE);
                } else {
                    return PreferencesSshKeyInfoFactory.createPreferencesSshKeyInfo(this, CorePreferenceKey.PREF_SSH_USE_CUSTOM_KEY,
                            CorePreferenceKey.PREF_SSH_PUBLIC_KEY_FILE,
                            CorePreferenceKey.PREF_SSH_PRIVATE_KEY_FILE,
                            CorePreferenceKey.PREF_SSH_PUBLIC_KEY,
                            CorePreferenceKey.PREF_SSH_PRIVATE_KEY);
                }
            } catch (InvalidSshKeyInfoException ex) {
                LOG.warn("Could not create a valid SshKeyInfo from the preferences", ex);
                return null;
            }
        } else
            return null;
    }
}
